@@ -5,6 +5,7 @@ synchronous and asynchronous operations.
 """
 
 import asyncio
+import base64
 import json
 import threading
 import time
@@ -22,6 +23,7 @@ from openai.types.chat import (
     ChatCompletionMessageToolCall,
 )
 from openai.types.chat.chat_completion_chunk import Choice, ChoiceDeltaToolCall
+from openai.types.responses import ParsedResponse
 
 from llmax.external_clients.clients import Client, get_aclient, get_client
 from llmax.messages import Messages
@@ -113,12 +115,40 @@ class MultiAIClient:
             self._aclients[model] = get_aclient(self.deployments[model])
         return self._aclients[model]
 
+    def clean_kwargs(
+        self,
+        kwargs: dict[str, Any],
+        deployment: Deployment,
+    ) -> dict[str, Any]:
+        """Clean kwargs to avoid errors."""
+        if "temperature" in kwargs and deployment.model in {"o3-mini", "o3-mini-high"}:
+            logger.warning("Temperature is not supported for this model.")
+            kwargs.pop("temperature")
+        if "text_format" in kwargs and deployment.model not in {
+            "gpt'4o",
+            "gpt-4o-mini",
+            "gpt-4.1",
+            "gpt-4.1-mini",
+            "gpt-5",
+            "gpt-5-mini",
+            "gpt-5-turbo",
+        }:
+            logger.warning("Text format is not supported for this model.")
+            kwargs.pop("text_format")
+        if "text_format" in kwargs and deployment.api_version < "2025-03-01-preview":
+            logger.warning("Text format is not supported for this API version.")
+            kwargs.pop("text_format")
+        if "reasoning_effort" in kwargs and deployment.model == "o3-mini-high":
+            logger.warning("Reasoning effort is not supported for this model.")
+            kwargs.pop("reasoning_effort")
+        return kwargs
+
     def _create_chat(
         self,
         messages: Messages,
         model: Model,
         **kwargs: Any,
-    ) -> ChatCompletion:
+    ) -> ChatCompletion | ParsedResponse[Any]:
         """Synchronously creates chat completions for the given messages and model.
 
         Args:
@@ -132,6 +162,15 @@ class MultiAIClient:
         client = self.client(model)
         deployment = self.deployments[model]
 
+        kwargs = self.clean_kwargs(kwargs, deployment)
+
+        if "text_format" in kwargs:
+            return client.responses.parse(
+                input=messages,
+                model=deployment.deployment_name,
+                **kwargs,
+            )
+
         return client.chat.completions.create(
             messages=messages,
             model=deployment.deployment_name,
@@ -143,7 +182,7 @@ class MultiAIClient:
         messages: Messages,
         model: Model,
         **kwargs: Any,
-    ) -> ChatCompletion:
+    ) -> ChatCompletion | ParsedResponse[Any]:
         """Asynchronously creates chat completions for the given messages and model.
 
         Args:
@@ -155,12 +194,22 @@ class MultiAIClient:
             ChatCompletion: The completion response from the API.
         """
         aclient = self.aclient(model)
-        result = await aclient.chat.completions.create(
+        deployment = self.deployments[model]
+
+        kwargs = self.clean_kwargs(kwargs, deployment)
+
+        if "text_format" in kwargs:
+            return await aclient.responses.parse(
+                input=messages,
+                model=deployment.deployment_name,
+                **kwargs,
+            )
+
+        return await aclient.chat.completions.create(
             messages=messages,
-            model=self.deployments[model].deployment_name,
+            model=deployment.deployment_name,
             **kwargs,
         )
-        return result
 
     def invoke(
         self,
@@ -170,7 +219,7 @@ class MultiAIClient:
         delay: float = 0.0,
         tries: int = 1,
         **kwargs: Any,
-    ) -> ChatCompletion:
+    ) -> ChatCompletion | ParsedResponse[Any]:
         """Synchronously invokes the API to get chat completions, tracking usage.
 
         Args:
@@ -186,15 +235,13 @@ class MultiAIClient:
         """
         start_time = time.time()
         operation: str = kwargs.pop("operation", "")
-        if "temperature" in kwargs and model in {"o3-mini", "o3-mini-high"}:
-            kwargs.pop("temperature")
         if system:
             messages = add_system_message(
                 messages=messages,
                 model=model,
                 system=system,
             )
-        response: ChatCompletion | None = None
+        response: ChatCompletion | ParsedResponse[Any] | None = None
         for _ in range(tries):
             try:
                 response = self._create_chat(messages, model, **kwargs, stream=False)
@@ -212,7 +259,16 @@ class MultiAIClient:
             message = "No usage for this request"
             raise ValueError(message)
         deployment = self.deployments[model]
-        usage = ModelUsage(deployment, self._increment_usage, response.usage)
+        response_usage = (
+            response.usage
+            if isinstance(response, ChatCompletion)
+            else CompletionUsage(
+                completion_tokens=response.usage.output_tokens,
+                prompt_tokens=response.usage.input_tokens,
+                total_tokens=response.usage.total_tokens,
+            )
+        )
+        usage = ModelUsage(deployment, self._increment_usage, response_usage)
 
         cost = usage.apply(
             operation=operation,
@@ -254,7 +310,11 @@ class MultiAIClient:
             tries=tries,
             **kwargs,
         )
-        return response.choices[0].message.content if response.choices else None
+        if isinstance(response, ChatCompletion):
+            return response.choices[0].message.content if response.choices else None
+        if not response:
+            return None
+        return response.output[0].content[0].text if response.output else None
 
     async def ainvoke(
         self,
@@ -264,7 +324,7 @@ class MultiAIClient:
         delay: float = 0.0,
         tries: int = 1,
         **kwargs: Any,
-    ) -> ChatCompletion:
+    ) -> ChatCompletion | ParsedResponse[Any]:
         """Asynchronously invokes the API to get chat completions, tracking usage.
 
         Args:
@@ -280,8 +340,6 @@ class MultiAIClient:
         """
         start_time = time.time()
         operation: str = kwargs.pop("operation", "")
-        if "temperature" in kwargs and model in {"o3-mini", "o3-mini-high"}:
-            kwargs.pop("temperature")
         if system:
             messages = add_system_message(
                 messages=messages,
@@ -289,7 +347,7 @@ class MultiAIClient:
                 system=system,
             )
 
-        response: ChatCompletion | None = None
+        response: ChatCompletion | ParsedResponse[Any] | None = None
         for _ in range(tries):
             try:
                 response = await self._acreate_chat(
@@ -312,7 +370,16 @@ class MultiAIClient:
             message = "No usage for this request"
             raise ValueError(message)
         deployment = self.deployments[model]
-        usage = ModelUsage(deployment, self._increment_usage, response.usage)
+        response_usage = (
+            response.usage
+            if isinstance(response, ChatCompletion)
+            else CompletionUsage(
+                completion_tokens=response.usage.output_tokens,
+                prompt_tokens=response.usage.input_tokens,
+                total_tokens=response.usage.total_tokens,
+            )
+        )
+        usage = ModelUsage(deployment, self._increment_usage, response_usage)
 
         cost = usage.apply(
             operation=operation,
@@ -354,7 +421,9 @@ class MultiAIClient:
             system=system,
             **kwargs,
         )
-        return response.choices[0].message.content if response.choices else None
+        if isinstance(response, ChatCompletion):
+            return response.choices[0].message.content if response.choices else None
+        return response.output[0].content[0].text if response.output else None
 
     async def ainvoke_get_tools(
         self,
@@ -445,18 +514,18 @@ class MultiAIClient:
                 )
                 messages.append({"role": "assistant", "content": output_str})
 
+            tool_coros = []
             for tool in final_tool_calls:
                 function_name = tool.function.name
                 function_args = tool.function.arguments
                 logger.info(
                     f"Tool called for function `{function_name}` with the args `{function_args}`",
                 )
+                tool_coros.append(execute_tools(function_name, function_args))
 
-                resultat = await execute_tools(
-                    function_name,
-                    function_args,
-                )
+            results = await asyncio.gather(*tool_coros)
 
+            for tool, resultat in zip(final_tool_calls, results):
                 messages.append(parse_tool_call(tool, model))
                 messages.append(
                     {
@@ -489,8 +558,6 @@ class MultiAIClient:
         start = time.time()
         ttft = None
         operation: str = kwargs.pop("operation", "")
-        if "temperature" in kwargs and model in {"o3-mini", "o3-mini-high"}:
-            kwargs.pop("temperature")
         if system:
             messages = add_system_message(
                 messages=messages,
@@ -749,32 +816,40 @@ class MultiAIClient:
             if output_str:
                 messages.append({"role": "assistant", "content": output_str})
 
-            for tool in final_tool_calls.values():
-                if tool.function is None:
-                    continue
+            queue = asyncio.Queue()
+
+            async def run_tool(tool: ChoiceDeltaToolCall) -> None:
+                tool_id = tool.id
                 function_name = tool.function.name
                 function_args = tool.function.arguments
-                tool_id = tool.id
-                if function_args is None or function_name is None or tool_id is None:
+                if not all([tool_id, function_name, function_args]):
+                    return
+                logger.info(
+                    f"Tool called for function `{function_name}` with args `{function_args}`",
+                )
+                async for res in execute_tools(function_name, function_args, tool_id):
+                    await queue.put((tool, res))
+                await queue.put((tool, None))
+
+            tasks = [
+                asyncio.create_task(run_tool(tool))
+                for tool in final_tool_calls.values()
+            ]
+            finished_tools = 0
+
+            while finished_tools < len(tasks):
+                tool, res = await queue.get()
+                if res is None:
+                    finished_tools += 1
                     continue
 
-                logger.info(
-                    f"Tool called for function `{function_name}` with the args `{function_args}`"
-                )
-
-                tool_result = None
-                tool_retrigger = False
-
-                async for res in execute_tools(function_name, function_args, tool_id):
-                    if isinstance(res, ToolItemContent):
-                        yield res.content
-                    else:
-                        tool_result, tool_retrigger = res.output, res.redo
-
-                if not tool_retrigger:
-                    retrigger_stream = False
-
-                update_messages_tools(tool, tool_result, messages, model)
+                if isinstance(res, ToolItemContent):
+                    yield res.content
+                else:
+                    tool_result, tool_retrigger = res.output, res.redo
+                    update_messages_tools(tool, tool_result, messages, model)
+                    if not tool_retrigger:
+                        retrigger_stream = False
 
             for res in fake_llm("\n", stream=False):
                 yield res
@@ -877,6 +952,7 @@ class MultiAIClient:
             file: The audio data to process.
             model: The model to use for processing the audio.
             response_format: for gpt-4o-transcribe, you cannot pass verbose_json.
+            duration: duration of the audio, if known.
             kwargs: Additional arguments to pass to the API.
 
         Returns:
@@ -918,15 +994,15 @@ class MultiAIClient:
 
         return response
 
-    def text_to_image(
+    async def text_to_image(
         self,
         model: Model,
         prompt: str,
-        size: Literal["1024x1024", "1024x1792", "1792x1024"] = "1024x1024",
-        quality: Literal["standard", "hd"] = "standard",
+        size: Literal["1024x1024", "1024x1536", "1536x1024"] = "1024x1024",
+        quality: Literal["low", "medium", "high", "auto"] = "medium",
         n: int = 1,
         **kwargs: Any,
-    ) -> str:
+    ) -> bytes:
         """Generate images from a text prompt using the specified model.
 
         Parameters:
@@ -949,12 +1025,10 @@ class MultiAIClient:
         """
         start = time.time()
         operation: str = kwargs.pop("operation", "")
-        if "temperature" in kwargs and model in {"o3-mini", "o3-mini-high"}:
-            kwargs.pop("temperature")
-        client = self.client(model)
+        client = self.aclient(model)
         deployment = self.deployments[model]
 
-        response = client.images.generate(
+        response = await client.images.generate(
             model=deployment.deployment_name,
             prompt=prompt,
             size=size,
@@ -978,7 +1052,8 @@ class MultiAIClient:
         self.total_usage += cost
         self.usages.append(usage)
 
-        return response.data[0].url
+        image_base64 = response.data[0].b64_json
+        return base64.b64decode(image_base64)
 
     def text_to_speech(
         self,
